#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/param.h>
#include "hxd.h"

#if defined(CONFIG_EXEC)

static void
exec_close (int fd)
{
	close(fd);
	memset(&hxd_files[fd], 0, sizeof(struct hxd_file));
	FD_CLR(fd, &hxd_rfds);
	if (high_fd == fd) {
		for (fd--; fd && !FD_ISSET(fd, &hxd_rfds); fd--)
			;
		high_fd = fd;
	}
}

static void
exec_ready_read (int fd)
{
	ssize_t r;
	u_int8_t buf[16384];

	r = read(fd, buf, sizeof(buf));
	if (r <= 0)
		exec_close(fd);
	else {
		hlwrite(hxd_files[fd].conn.htlc, HTLS_HDR_CHAT, 0, 1,
			HTLS_DATA_CHAT, r, buf);
	}
}

#if 0 /* add somethin like http post here?? */
static void
exec_ready_write (int fd)
{
}
#endif

static void
cmd_exec (struct htlc_conn *htlc, char *command)
{
	int argc, pfds[2];
	char *argv[32], *p, *pii, *thisarg, cmdpath[MAXPATHLEN];

	while ((p = strstr(command, "../"))) {
		for (pii = p; *pii; pii++)
			*pii = *(pii + 3);
	}

	for (argc = 0, thisarg = p = command; *p && argc < 30; p++) {
		if (*p == ' ') {
			*p = 0;
			argv[argc++] = thisarg;
			thisarg = p + 1;
		}
	}
	if (thisarg != p)
		argv[argc++] = thisarg;
	argv[argc] = 0;
	snprintf(cmdpath, sizeof(cmdpath), "%s/%s", hxd_cfg.paths.exec, argv[0]);

#if 0
	hxd_log("%s execs %s", htlc->name, cmdpath);
#endif
	if (pipe(pfds)) {
		hxd_log("cmd_exec: pipe: %s", strerror(errno));
		return;
	}
	if (pfds[0] >= hxd_open_max) {
		hxd_log("%s:%d: %d >= hxd_open_max (%d)", __FILE__, __LINE__, pfds[0], hxd_open_max);
		close(pfds[0]);
		close(pfds[1]);
		return;
	}
	switch (fork()) {
		case -1:
			hxd_log("cmd_exec: fork: %s", strerror(errno));
			close(pfds[0]);
			close(pfds[1]);
			return;
		case 0:
			if (dup2(pfds[1], 1) == -1 || dup2(pfds[1], 2) == -1) {
				hxd_log("cmd_exec: dup2: %s", strerror(errno));
				_exit(1);
			}
			close(0);
#if 0
			{
				int i;

				fprintf(stderr, "executing");
				for (i = 0; i < argc; i++)
					fprintf(stderr, " %s", argv[i]);
			}
#endif
			{
				char *envp[2], rootdir[MAXPATHLEN + 16];

				snprintf(rootdir, sizeof(rootdir), "ROOTDIR=%s", htlc->rootdir);
				envp[0] = rootdir;
				envp[1] = 0;
				execve(cmdpath, argv, envp);
			}
			fprintf(stderr, "\r%s: %s", argv[0], strerror(errno));
			_exit(1);
		default:
			close(pfds[1]);
			hxd_files[pfds[0]].conn.htlc = htlc;
			hxd_files[pfds[0]].ready_read = exec_ready_read;
			FD_SET(pfds[0], &hxd_rfds);
			if (high_fd < pfds[0])
				high_fd = pfds[0];
			break;
	}
}
#endif /* ndef CONFIG_EXEC */

void
toggle_away (struct htlc_conn *htlc)
{
	u_int16_t uid, icon16, color;
	struct htlc_conn *htlcp;

	htlc->color = htlc->color & 1 ? htlc->color-1 : htlc->color+1;
	if (!htlc->flags.visible)
		return;
#if defined(CONFIG_SQL)
	sql_modify_user(htlc->name, htlc->icon, htlc->color, htlc->uid);
#endif
	uid = htons(mangle_uid(htlc));
	icon16 = htons(htlc->icon);
	color = htons(htlc->color);
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
		if (!htlcp->access_extra.user_getlist)
			continue;
		hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
			HTLS_DATA_UID, sizeof(uid), &uid,
			HTLS_DATA_ICON, sizeof(icon16), &icon16,
			HTLS_DATA_COLOUR, sizeof(color), &color,
			HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
	}
}

static void
cmd_broadcast (struct htlc_conn *htlc, char *chatbuf)
{
	u_int16_t style, len;
	struct htlc_conn *htlcp;

	if (!htlc->access.can_broadcast)
		return;

	style = htons(1);
	len = strlen(chatbuf);
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next)
		hlwrite(htlcp, HTLS_HDR_MSG, 0, 2,
			HTLS_DATA_STYLE, 2, &style,
			HTLS_DATA_MSG, len, chatbuf);
}

static void
cmd_visible (struct htlc_conn *htlc)
{
	u_int16_t uid16, icon16, color;
	struct htlc_conn *htlcp;

	if (!htlc->access_extra.user_visibility)
		return;

	htlc->flags.visible = !htlc->flags.visible;
	if (!htlc->flags.visible) {
		uid16 = htons(mangle_uid(htlc));
		for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
			if (!htlcp->access_extra.user_getlist)
				continue;
			hlwrite(htlcp, HTLS_HDR_USER_PART, 0, 1,
				HTLS_DATA_UID, sizeof(uid16), &uid16);
		}
	} else {
		uid16 = htons(mangle_uid(htlc));
		icon16 = htons(htlc->icon);
		color = htons(htlc->color);
		for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
			if (!htlcp->access_extra.user_getlist)
				continue;
			hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
				HTLS_DATA_UID, sizeof(uid16), &uid16,
				HTLS_DATA_ICON, sizeof(icon16), &icon16,
				HTLS_DATA_COLOUR, sizeof(color), &color,
				HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
		}
	}
}

static void
cmd_color (struct htlc_conn *htlc, char *chatbuf)
{
	u_int16_t uid16, icon16, color = htlc->color;
	struct htlc_conn *htlcp;

	if (!htlc->access_extra.user_color)
		return;

	htlc->color = strtoul(chatbuf, 0, 0);
#if defined(CONFIG_SQL)
	sql_modify_user(htlc->name, htlc->icon, htlc->color, htlc->uid);
#endif
	if (htlc->color != color && htlc->flags.visible) {
		uid16 = htons(mangle_uid(htlc));
		icon16 = htons(htlc->icon);
		color = htons(htlc->color);
		for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
			if (!htlcp->access_extra.user_getlist)
				continue;
			hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
				HTLS_DATA_UID, sizeof(uid16), &uid16,
				HTLS_DATA_ICON, sizeof(icon16), &icon16,
				HTLS_DATA_COLOUR, sizeof(color), &color,
				HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
		}
	}
}

/* chatbuf points to the auto array in rcv_chat + 1 */
void
command_chat (struct htlc_conn *htlc, char *chatbuf)
{
	switch (chatbuf[0]) {
		case 'a':
			if (!strncmp(chatbuf, "away", 4)) {
				if (htlc->flags.away == AWAY_INTERRUPTED)
					break;
				toggle_away(htlc);
				if (!htlc->flags.away)
					htlc->flags.away = AWAY_PERM;
				else
					htlc->flags.away = 0;
				if (hxd_cfg.options.away_time) {
					timer_delete_ptr(htlc);
					if (!htlc->flags.away)
						timer_add_secs(hxd_cfg.options.away_time, away_timer, htlc);
				}
			}
			break;
		case 'b':
			if (!strncmp(chatbuf, "broadcast ", 10) && chatbuf[10])
				cmd_broadcast(htlc, &chatbuf[10]);
		case 'c':
			if (!strncmp(chatbuf, "color ", 6) && chatbuf[6])
				cmd_color(htlc, &chatbuf[6]);
#if defined(CONFIG_EXEC)
		case 'e':
			if (!strncmp(chatbuf, "exec ", 5) && chatbuf[5])
				cmd_exec(htlc, &chatbuf[5]);
			break;
		case 'f':
			if (!strncmp(chatbuf, "find", 4))
				cmd_exec(htlc, chatbuf);
			break;
#endif
		case 'g':
			if (!strncmp(chatbuf, "g0away", 6))
				cmd_visible(htlc);
			break;
		case 'v':
			if (!strncmp(chatbuf, "visible", 7))
				cmd_visible(htlc);
			break;
#if XMALLOC_DEBUG
		case 'm':
			if (!strncmp(chatbuf, "maltbl", 6) && htlc->access_extra.debug) {
				extern void DTBLWRITE (void);
				hxd_log("%s: writing maltbl", htlc->login);
				DTBLWRITE();
			}
			break;
#endif
		default:
			break;
	}
}
