#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include "hxd.h"
#include "xmalloc.h"
#ifdef SOCKS
#include "socks.h"
#endif

#if defined(ONLY_TRACKER)
#undef CONFIG_TRACKER_REGISTER
#elif defined(ONLY_SERVER)
#undef CONFIG_TRACKER_SERVER
#endif

#if defined(CONFIG_TRACKER_SERVER)
struct server {
	struct sockaddr_in sockaddr;
	u_int16_t port;
	u_int16_t nusers;
	u_int32_t id;
	u_int8_t nlen, name[31];
	u_int8_t dlen, desc[255];
	u_int32_t clock;
};

static struct server *servers = 0;
static u_int16_t nservers = 0;
static u_int8_t *__buf = 0;
static u_int32_t __buf_len = 0;

#define MAX_SERVERS	1000

static int
tracker_timer (void)
{
	unsigned int i;

	for (i = 0; i < nservers; ) {
		if (servers[i].clock == 2) {
			memcpy(&servers[i], &servers[i + 1], (nservers - i) * sizeof(struct server));
			nservers--;
		} else {
			servers[i].clock++;
			i++;
		}
	}

	return 1;
}

static void
tracker_client_close (int fd)
{
	close(fd);
	FD_CLR(fd, &hxd_rfds);
	FD_CLR(fd, &hxd_wfds);
	if (hxd_files[fd].conn.htrk->out.buf)
		xfree(hxd_files[fd].conn.htrk->out.buf);
	xfree(hxd_files[fd].conn.htrk);
	memset(&hxd_files[fd], 0, sizeof(struct hxd_file));
}

static void
tracker_client_write (int fd)
{
	ssize_t r;
	struct htrk_conn *htrk = hxd_files[fd].conn.htrk;

	r = write(fd, &htrk->out.buf[htrk->out.pos], htrk->out.len);
	if (r == 0 || (r < 0 && errno != EWOULDBLOCK && errno != EINTR)) {
		tracker_client_close(fd);
	} else {
		htrk->out.pos += r;
		htrk->out.len -= r;
		if (!htrk->out.len)
#if 0
			tracker_client_close(fd);
#else
		{
			FD_CLR(fd, &hxd_wfds);
			FD_SET(fd, &hxd_rfds);
		}
#endif

	}
}

static void
tracker_client_read (int fd)
{
	int r;
	u_int8_t buf[6];

	r = read(fd, buf, sizeof(buf));
	if (r == 0 || (r < 0 && errno != EWOULDBLOCK && errno != EINTR)) {
		tracker_client_close(fd);
	} else {
		FD_CLR(fd, &hxd_rfds);
		FD_SET(fd, &hxd_wfds);
		r = write(fd, HTRK_MAGIC, HTRK_MAGIC_LEN);
		if (r != HTRK_MAGIC_LEN)
			tracker_client_close(fd);
	}
}

static void
tracker_tcp_ready_read (int fd)
{
	struct sockaddr_in saddr;
	int siz = sizeof(saddr);
	int s;
	unsigned int i;
	u_int8_t *buf = __buf;
	u_int32_t buf_len = __buf_len, pos = 8, thispos;
	u_int16_t thisnserv;
	struct htrk_conn *htrk;

	s = accept(fd, (struct sockaddr *)&saddr, &siz);
	if (s < 0) {
		hxd_log("accept: %s", strerror(errno));
		return;
	}
	if (s >= hxd_open_max) {
		hxd_log("%s:%d: %d >= hxd_open_max (%d)", __FILE__, __LINE__, s, hxd_open_max);
		close(s);
		return;
	}
	fd_closeonexec(s, 1);
	fd_blocking(s, 0);
	if (high_fd < s)
		high_fd = s;

#if defined(SO_LINGER)
	{
		struct linger lingivitis;

		lingivitis.l_onoff = 1;
		lingivitis.l_linger = 1000;
		setsockopt(s, SOL_SOCKET, SO_LINGER, (char *)&lingivitis, sizeof(lingivitis));
	}
#endif

	htrk = xmalloc(sizeof(struct htrk_conn));
	memset(htrk, 0, sizeof(struct htrk_conn));

	hxd_files[s].ready_read = tracker_client_read;
	hxd_files[s].ready_write = tracker_client_write;
	hxd_files[s].conn.htrk = htrk;
	FD_SET(s, &hxd_rfds);

	thispos = 0;
	thisnserv = 0;
	for (i = 0; i < nservers; i++) {
		if ((pos + (12 + servers[i].nlen + servers[i].dlen)) - thispos > 0x1fff) {
			buf[thispos] = 0;
			buf[thispos + 1] = 1;
			S16HTON(pos - (thispos + 4), &buf[thispos + 2]);
			S16HTON(nservers, &buf[thispos + 4]);
			S16HTON(thisnserv, &buf[thispos + 6]);
			if (i + 1 != nservers) {
				thispos = pos;
				thisnserv = 0;
			}
			pos += 8;
		}
		if (buf_len < pos + (12 + servers[i].nlen + servers[i].dlen)) {
			buf_len = pos + 12 + servers[i].nlen + servers[i].dlen;
			buf = xrealloc(buf, buf_len);
		}
		thisnserv++;
		memcpy(&buf[pos], &servers[i].sockaddr.sin_addr.s_addr, 4);
		pos += 4;
		S16HTON(servers[i].port, &buf[pos]);
		pos += 2;
		S16HTON(servers[i].nusers, &buf[pos]);
		pos += 2;
		buf[pos++] = 0;
		buf[pos++] = 0;
		buf[pos++] = servers[i].nlen;
		memcpy(&buf[pos], servers[i].name, servers[i].nlen);
		pos += servers[i].nlen;
		buf[pos++] = servers[i].dlen;
		memcpy(&buf[pos], servers[i].desc, servers[i].dlen);
		pos += servers[i].dlen;
	}
	buf[thispos] = 0;
	buf[thispos + 1] = 1;
	S16HTON(pos - (thispos + 4), &buf[thispos + 2]);
	S16HTON(nservers, &buf[thispos + 4]);
	S16HTON(thisnserv, &buf[thispos + 6]);
	qbuf_add(&htrk->out, buf, pos);

	__buf = buf;
	__buf_len = buf_len;
}

static unsigned int nbanned;
static u_int32_t *banned_addresses;
static u_int8_t *banned_wildcards;

static int
banned (u_int32_t addr)
{
	u_int32_t *aend = banned_addresses + nbanned;
	u_int32_t *ap, a;
	u_int8_t *wp, w;

	for (ap = banned_addresses, wp = banned_wildcards; ap < aend; ap++, wp++) {
		a = *ap;
		w = *wp;
		if ((a & 0xff000000) != (addr & 0xff000000) && !(w & 8))
			continue;
		if ((a & 0xff0000) != (addr & 0xff0000) && !(w & 4))
			continue;
		if ((a & 0xff00) != (addr & 0xff00) && !(w & 2))
			continue;
		if ((a & 0xff) != (addr & 0xff) && !(w & 1))
			continue;
		return 1;
	}

	return 0;
}

void
tracker_read_banlist (void)
{
	char buf[1024], *p, *nlp, *dp;
	u_int32_t addr;
	u_int8_t wild;
	unsigned int shift, idx = 0;
	ssize_t r;
	int fd;

	fd = open(hxd_cfg.paths.tracker_banlist, O_RDONLY);
	if (fd < 0)
		return;
	for (;;) {
		r = read(fd, buf, sizeof(buf)-1);
		if (r <= 0)
			break;
		buf[r] = 0;
		for (p = buf-1; p < buf+r; p = nlp) {
			p++;
			nlp = strchr(p, '\n');
			if (nlp)
				*nlp = 0;
			if (*p == '#')
				goto may_continue;
			addr = 0;
			wild = 0;
			shift = 3;
			for (dp = p; dp <= nlp; dp++) {
				if (*dp == '.' || !*dp) {
					*dp = 0;
					if (*p == '*')
						wild |= (1<<shift);
					else
						addr |= (atoi(p) & 0xff) << (shift*8);
					p = dp+1;
					if (!shift)
						break;
					shift--;
				}
			}
			if (shift != 0)
				goto may_continue;
			banned_addresses = xrealloc(banned_addresses, sizeof(u_int32_t)*(idx+1));
			banned_wildcards = xrealloc(banned_wildcards, sizeof(u_int8_t)*(idx+1));
			banned_addresses[idx] = addr;
			banned_wildcards[idx] = wild;
			idx++;
may_continue:
			if (!nlp)
				break;
		}
	}
	close(fd);
	nbanned = idx;
}

static void
tracker_udp_ready_read (int fd)
{
	struct sockaddr_in saddr;
	int siz = sizeof(saddr);
	ssize_t r;
	u_int8_t buf[1024];
	int totlen, nlen, dlen, plen, x = 0;
	char *tracker_password;
	int tracker_passlen;

	memset(buf, 0, sizeof(buf));
	r = recvfrom(fd, buf, sizeof(buf), 0, (struct sockaddr *)&saddr, &siz);
	if (r < 0) {
		hxd_log("recvfrom: %s", strerror(errno));
		return;
	}

	if (banned(ntohl(saddr.sin_addr.s_addr)))
		return;

	if (SIZEOF_HTRK_HDR > r)
		return;

	nlen = buf[12];
	if (nlen > 31)
		nlen = 31;
	dlen = buf[13 + nlen];
	totlen = SIZEOF_HTRK_HDR + 2 + nlen + dlen;
	if (totlen > r)
		return;

	tracker_password = hxd_cfg.tracker.password;
	if (tracker_password) {
		tracker_passlen = strlen(tracker_password);
		if (totlen == r || totlen+1 == r)
			return;
		plen = buf[totlen];
		if (plen != tracker_passlen || memcmp(&buf[totlen+1], tracker_password, plen))
			return;
	}

	for (x = 0; x < nservers; x++) {
		if (servers[x].sockaddr.sin_addr.s_addr == saddr.sin_addr.s_addr
		    && servers[x].sockaddr.sin_port == saddr.sin_port)
			goto hi_again;
	}
	if (x >= MAX_SERVERS)
		return;
	nservers++;
hi_again:
	memcpy(&servers[x].sockaddr, &saddr, sizeof(servers[x].sockaddr));
	L16NTOH(servers[x].port, &buf[2]);
	L16NTOH(servers[x].nusers, &buf[4]);
	L32NTOH(servers[x].id, &buf[8]);
	servers[x].nlen = nlen;
	servers[x].dlen = dlen;
	memcpy(servers[x].name, &buf[13], nlen);
	memcpy(servers[x].desc, &buf[14 + nlen], dlen);
	servers[x].clock = 0;
}

static void
init_servers (void)
{
#if 0
	u_int32_t addr = 0x7f000001;
	u_int16_t port = 5500;
	u_int16_t nusers = 1;
	u_int32_t id = 31337;
	unsigned int i;

	for (i = 0; i < 10; i++) {
		servers[i].sockaddr.sin_addr.s_addr = htonl(addr++);
		servers[i].port = port++;
		servers[i].nusers = nusers++;
		servers[i].id = id++;
		servers[i].nlen = sprintf(servers[i].name, "haxor %d", (rand()>>16)%10);
		servers[i].dlen = sprintf(servers[i].desc, "i am leet %d++", (rand()>>16)%100);
		servers[i].clock = 0;
		nservers++;
	}
#endif
}

void
tracker_server_init (void)
{
	int t_sock, u_sock;
	int x;
	struct sockaddr_in saddr;
	int i;
	char *host;

	for (i = 0; (host = hxd_cfg.options.addresses[i]); i++) {
		if (!inet_aton(host, &saddr.sin_addr)) {
			struct hostent *he = gethostbyname(host);
			if (he)
				memcpy(&saddr.sin_addr, he->h_addr_list[0],
				       (size_t)he->h_length > sizeof(saddr.sin_addr) ?
				       sizeof(saddr.sin_addr) : (size_t)he->h_length);
			else {
				hxd_log("%s:%d: could not resolve hostname %s", __FILE__, __LINE__, host);
				exit(1);
			}
		}

		u_sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (u_sock < 0) {
			hxd_log("%s:%d: socket: %s", __FILE__, __LINE__, strerror(errno));
			exit(1);
		}
		if (u_sock >= hxd_open_max) {
			hxd_log("%s:%d: %d >= hxd_open_max (%d)", __FILE__, __LINE__, u_sock, hxd_open_max);
			close(u_sock);
			exit(1);
		}
		saddr.sin_family = AF_INET;
		saddr.sin_port = htons(HTRK_UDPPORT);
		if (bind(u_sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
			hxd_log("%s:%d: bind: %s", __FILE__, __LINE__, strerror(errno));
			exit(1);
		}
		fd_closeonexec(u_sock, 1);
		fd_blocking(u_sock, 0);
		hxd_files[u_sock].ready_read = tracker_udp_ready_read;
		FD_SET(u_sock, &hxd_rfds);
		if (high_fd < u_sock)
			high_fd = u_sock;

		t_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (t_sock < 0) {
			hxd_log("%s:%d: socket: %s", __FILE__, __LINE__, strerror(errno));
			exit(1);
		}
		if (t_sock >= hxd_open_max) {
			hxd_log("%s:%d: %d >= hxd_open_max (%d)", __FILE__, __LINE__, t_sock, hxd_open_max);
			close(t_sock);
			exit(1);
		}
		x = 1;
		setsockopt(t_sock, SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x));
		saddr.sin_family = AF_INET;
		saddr.sin_port = htons(HTRK_TCPPORT);
		if (bind(t_sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
			hxd_log("%s:%d: bind: %s", __FILE__, __LINE__, strerror(errno));
			exit(1);
		}
		if (listen(t_sock, 5) < 0) {
			hxd_log("%s:%d: listen: %s", __FILE__, __LINE__, strerror(errno));
			exit(1);
		}
		fd_closeonexec(t_sock, 1);
		fd_blocking(t_sock, 0);
		hxd_files[t_sock].ready_read = tracker_tcp_ready_read;
		FD_SET(t_sock, &hxd_rfds);
		if (high_fd < t_sock)
			high_fd = t_sock;
	}

	servers = xmalloc(sizeof(struct server) * MAX_SERVERS);
	memset(servers, 0, sizeof(struct server) * MAX_SERVERS);

	init_servers();

	tracker_read_banlist();

	__buf_len = 8;
	__buf = xmalloc(__buf_len);

	timer_add_secs(hxd_cfg.tracker.interval, tracker_timer, 0);
}
#endif /* ndef CONFIG_TRACKER_SERVER */

#if defined(CONFIG_TRACKER_REGISTER)

struct tracker {
	struct sockaddr_in sockaddr;
	u_int32_t id;
	char plen, password[31];
};

static int ntrackersocks = 0;
static int *tracker_udpsocks = 0;

static int ntrackers = 0;
static struct tracker *trackers = 0;

static u_int8_t tracker_buf[sizeof(struct htrk_hdr) + 512];
static u_int16_t tracker_buf_len = 0;

int
return_num_users (void)
{
	struct htlc_conn *htlcp;
	int i = 0;
	
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next)
		i++;
	
	return i;
}

int
tracker_register_timer (void *__arg)
{
	struct htrk_hdr *pkt = (struct htrk_hdr *)tracker_buf;
	int i, j, len, plen;

	pkt->port = htons(hxd_cfg.options.htls_port);
	if (hxd_cfg.tracker.nusers >= 0)
		pkt->nusers = htons(hxd_cfg.tracker.nusers);
	else
		pkt->nusers = htons(return_num_users());

	for (i = 0; i < ntrackers; i++) {
		pkt->id = htonl(trackers[i].id);
		len = tracker_buf_len;
		plen = trackers[i].plen;
		if (plen) {
			tracker_buf[len++] = plen;
			memcpy(&tracker_buf[len], trackers[i].password, plen);
			len += plen;
		}
		for (j = 0; j < ntrackersocks; j++)
			(void)sendto(tracker_udpsocks[j], tracker_buf, len, 0,
				     (struct sockaddr *)&trackers[i].sockaddr, sizeof(struct sockaddr_in));
	}
	if (__arg)
		return 1;
	else
		return 0;
}

void
tracker_register_init (void)
{
	struct htrk_hdr *pkt = (struct htrk_hdr *)tracker_buf;
	struct sockaddr_in saddr;
	struct in_addr inaddr;
	u_int16_t len;
	u_int8_t nlen, dlen;
	int i;
	u_int32_t id;
	char *trp, *pass, *host;

	ntrackers = 0;
	if (!hxd_cfg.tracker.trackers) {
		if (trackers) {
			xfree(trackers);
			trackers = 0;
		}
		return;
	}
	for (i = 0; (trp = hxd_cfg.tracker.trackers[i]); i++) {
		host = strrchr(trp, '@');
		if (!host) {
			host = trp;
			id = 0;
			pass = 0;
		} else {
			*host = 0;
			host++;
			pass = strchr(trp, ':');
			if (pass)
				*pass = 0;
			id = strtoul(trp, 0, 0);
			if (pass) {
				*pass = 0;
				pass++;
			}
		}
		if (!inet_aton(host, &inaddr)) {
			struct hostent *he = gethostbyname(host);
			if (he)
				memcpy(&inaddr, he->h_addr_list[0],
				       (unsigned)he->h_length > sizeof(inaddr) ?
				       sizeof(inaddr) : (unsigned)he->h_length);
			else {
				hxd_log("could not resolve tracker hostname %s", host);
				/* exit(1); */
				continue;
			}
		}
		trackers = xrealloc(trackers, sizeof(struct tracker)*(ntrackers+1));
		trackers[ntrackers].sockaddr.sin_addr = inaddr;
		trackers[ntrackers].sockaddr.sin_port = htons(HTRK_UDPPORT);
		trackers[ntrackers].sockaddr.sin_family = AF_INET;
		trackers[ntrackers].id = id;
		if (pass) {
			int plen = strlen(pass);
			if (plen > 30)
				plen = 30;
			memcpy(trackers[ntrackers].password, pass, plen);
			trackers[ntrackers].password[plen] = 0;
			trackers[ntrackers].plen = plen;
		} else {
			trackers[ntrackers].plen = 0;
			trackers[ntrackers].password[0] = 0;
		}
		ntrackers++;
	}
	if (!ntrackers)
		return;

	for (i = 0; i < ntrackersocks; i++)
		close(tracker_udpsocks[i]);
	ntrackersocks = 0;

	for (i = 0; (host = hxd_cfg.options.addresses[i]); i++) {
		if (!inet_aton(host, &saddr.sin_addr)) {
			struct hostent *he = gethostbyname(host);
			if (he) {
				memcpy(&saddr.sin_addr, he->h_addr_list[0],
				       (size_t)he->h_length > sizeof(saddr.sin_addr) ?
				       sizeof(saddr.sin_addr) : (size_t)he->h_length);
			} else {
				hxd_log("%s:%d: could not resolve hostname %s", __FILE__, __LINE__, host);
				/* exit(1); */
				continue;
			}
		}

		tracker_udpsocks = xrealloc(tracker_udpsocks, sizeof(int)*(ntrackersocks+1));
		tracker_udpsocks[ntrackersocks] = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (tracker_udpsocks[ntrackersocks] < 0) {
			hxd_log("tracker_init: socket: %s", strerror(errno));
			/* exit(1); */
			continue;
		}
		saddr.sin_family = AF_INET;
		saddr.sin_port = 0; /* htons(32769); */
		if (bind(tracker_udpsocks[ntrackersocks], (struct sockaddr *)&saddr, sizeof(saddr))) {
			hxd_log("tracker_init: bind: %s", strerror(errno));
			close(tracker_udpsocks[ntrackersocks]);
			/* exit(1); */
			continue;
		}
		fd_closeonexec(tracker_udpsocks[ntrackersocks], 1);
		ntrackersocks++;
	}
	if (!ntrackersocks)
		return;

	len = sizeof(struct htrk_hdr);
	nlen = strlen(hxd_cfg.tracker.name);
	tracker_buf[len++] = nlen;
	memcpy(&tracker_buf[len], hxd_cfg.tracker.name, nlen);
	len += nlen;
	dlen = strlen(hxd_cfg.tracker.description);
	tracker_buf[len++] = dlen;
	memcpy(&tracker_buf[len], hxd_cfg.tracker.description, dlen);
	len += dlen;
	tracker_buf_len = len;
	pkt->version = htons(0x0001);
	pkt->__reserved0 = 0;
	pkt->id = 0;

	timer_add_secs(hxd_cfg.tracker.interval, tracker_register_timer, trackers);
}
#endif /* ndef CONFIG_TRACKER_REGISTER */
